import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

import 'apm_root_widget.dart';
import 'channel_info_manager.dart';

/// 以自定义WidgetsFlutterBinding形式使用
/// 一般为单独使用APM时使用的
class ApmWidgetsFlutterBinding extends WidgetsFlutterBinding
    with ApmBinding {
  static WidgetsBinding? ensureInitialized() {
    assert(WidgetsBinding.instance == null || WidgetsBinding.instance is ApmBinding, 'Pelease call runOnekitApp() first in main().');
    if (WidgetsBinding.instance == null) {
      ApmWidgetsFlutterBinding();
    }
    return WidgetsBinding.instance;
  }
}

/// 以自定义mixin的Binding形式使用
/// 一般为融合到OneKit时使用的
mixin ApmBinding on BindingBase, WidgetsBinding, ServicesBinding {
  @override
  void scheduleAttachRootWidget(Widget rootWidget) {
    // 在根节点套一层我们的widget
    super.scheduleAttachRootWidget(ApmRootWidget(
      child: rootWidget,
    ));
  }

  @override
  BinaryMessenger createBinaryMessenger() {
    return ApmBinaryMessenger(super.createBinaryMessenger());
  }
}

class ApmBinaryMessenger extends BinaryMessenger {
  final BinaryMessenger origin;

  ApmBinaryMessenger(this.origin);

  @visibleForTesting
  static ChannelInfoManager? mockChannelInfoManager;

  @override
  Future<void> handlePlatformMessage(String channel, ByteData? data,
      PlatformMessageResponseCallback? callback) async {
    final ChannelInfo? info =
        _parseChannelInfo(channel, data, InterceptType.PLATFORM_MESSAGE);
    final PlatformMessageResponseCallback wrapper = (ByteData? callbackData) {
      _parseResult(info, callbackData);
      callback?.call(callbackData);
    };

    return origin.handlePlatformMessage(channel, data, wrapper);
  }

  @override
  Future<ByteData?> send(String channel, ByteData? message) async {
    final ChannelInfo? info =
        _parseChannelInfo(channel, message, InterceptType.SEND);
    ByteData? result = await origin.send(channel, message);
    _parseResult(info, result);
    return result;
  }

  @override
  void setMessageHandler(
      String channel, Future<ByteData?>? Function(ByteData? message)? handler) {
    origin.setMessageHandler(channel, handler);
  }

  ChannelInfo? _parseChannelInfo(
      String name, ByteData? message, InterceptType interceptType) {
    if (message == null) {
      return null;
    }

    final ChannelInfo info = ChannelInfo(
      channelName: name,
      interceptType: interceptType,
      inputSize: message.lengthInBytes,
      startTimestamp: DateTime.now().millisecondsSinceEpoch,
    );
    dynamic codec = _getSystemChannelsCodec(info);
    if (codec != null) {
      info.identifyType = IdentityType.SYSTEM;
    } else {
      codec = const StandardMethodCodec();
      info.identifyType = IdentityType.USER;
    }
    if (codec is MethodCodec) {
      try {
        final MethodCall methodCall = codec.decodeMethodCall(message);
        info.methodName = methodCall.method;
      } catch (e) {
        if (e is FormatException && e.message == 'Invalid method call' && kDebugMode) {
          print('methodName is null, maybe is EventChannel call');
        }
      }
    }
    return info;
  }

  dynamic? _getSystemChannelsCodec(ChannelInfo info) {
    dynamic codec;
    if (info.channelName == SystemChannels.accessibility.name) {
      codec = SystemChannels.accessibility.codec;
    } else if (info.channelName == SystemChannels.keyEvent.name) {
      codec = SystemChannels.keyEvent.codec;
    } else if (info.channelName == SystemChannels.lifecycle.name) {
      codec = SystemChannels.lifecycle.codec;
    } else if (info.channelName == SystemChannels.mouseCursor.name) {
      codec = SystemChannels.mouseCursor.codec;
    } else if (info.channelName == SystemChannels.navigation.name) {
      codec = SystemChannels.navigation.codec;
    } else if (info.channelName == SystemChannels.platform.name) {
      codec = SystemChannels.platform.codec;
    } else if (info.channelName == SystemChannels.platform_views.name) {
      codec = SystemChannels.platform_views.codec;
    } else if (info.channelName == SystemChannels.restoration.name) {
      codec = SystemChannels.restoration.codec;
    } else if (info.channelName == SystemChannels.skia.name) {
      codec = SystemChannels.skia.codec;
    } else if (info.channelName == SystemChannels.system.name) {
      codec = SystemChannels.system.codec;
    } else if (info.channelName == SystemChannels.textInput.name) {
      codec = SystemChannels.textInput.codec;
    }
    return codec;
  }

  void _parseResult(ChannelInfo? info, ByteData? result) {
    if (info == null) {
      return;
    }
    info.outputSize = result?.lengthInBytes ?? 0;
    info.timeCost = DateTime.now().millisecondsSinceEpoch - info.startTimestamp;
    final ChannelInfoManager channelInfoManager =
        mockChannelInfoManager ?? ChannelInfoManager.instance;
    channelInfoManager.enqueue(info);
  }

  @override
  void setMockMessageHandler(
      String channel, Future<ByteData?>? Function(ByteData? message)? handler) {
    origin.setMockMessageHandler(channel, handler);
  }

  @override
  bool checkMessageHandler(String channel, MessageHandler? handler) {
    return origin.checkMessageHandler(channel, handler);
  }

  @override
  bool checkMockMessageHandler(String channel, MessageHandler? handler) {
    return origin.checkMockMessageHandler(channel, handler);
  }
}
