import 'dart:collection';
import 'dart:developer';
import 'package:ve_apm/ve_apm.dart';
import 'package:ve_apm/src/perf/app_context.dart';
import 'package:ve_apm/src/perf/fps_monitor.dart';
import 'package:ve_apm/src/perf/page_load_data_manager.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';

import 'apm_page_time_monitor.dart';

///
/// 在自研引擎里，使用匿名路由时，改过的dart编译会在路由里加上Widget文件在系统里的绝对路径名，
/// 但在官方引擎里，匿名路由我们toString获取到的是没有意义的，类似
/// :"MaterialPageRoute<dynamic>(RouteSettings(\"null\", null), animation: AnimationController#95029(⏭ 1.000; paused; for MaterialPageRoute<dynamic>(null)))"
///
class ApmNavigatorObserver extends NavigatorObserver {
  ApmNavigatorObserver._internal();

  static ApmNavigatorObserver? _instance;

  static ApmNavigatorObserver get instance =>
      _instance ??= ApmNavigatorObserver._internal();

  static const int maxRouteHistory = 15;
  Route<dynamic>? _currentRoute;
  ListQueue<String> routeHistory = ListQueue<String>(maxRouteHistory);

  @visibleForTesting
  FpsMonitor? mockFpsMonitor;
  @visibleForTesting
  APMPageTimeMonitor? mockApmPageTimeMonitor;
  @visibleForTesting
  PageLoadDataManager? mockPageLoadDataManager;
  @visibleForTesting
  AppContext? mockAppContext;

  @visibleForTesting
  void setMockObject({
    FpsMonitor? mockFpsMonitor,
    APMPageTimeMonitor? mockApmPageTimeMonitor,
    PageLoadDataManager? mockPageLoadDataManager,
    AppContext? mockAppContext,
  }) {
    this.mockFpsMonitor = mockFpsMonitor;
    this.mockApmPageTimeMonitor = mockApmPageTimeMonitor;
    this.mockPageLoadDataManager = mockPageLoadDataManager;
    this.mockAppContext = mockAppContext;
  }

  static String getRouteName(Route? route) {
    return route?.settings.name ?? UNKNOWN;
  }

  String getCurrentRouteName() {
    return getRouteName(_currentRoute);
  }

  ListQueue<String> getRecentRouteHistory() {
    return routeHistory;
  }

  @override
  void didPush(Route route, Route? previousRoute) {
    Timeline.startSync('self didPush');

    FlutterError.resetErrorCount();
    _currentRoute = route;
    _recordRoute(_RouteAction.PUSH, route, previousRoute);

    String page = getRouteName(route);
    String previousPage = getRouteName(previousRoute);
    // fps
    final FpsMonitor fpsMonitor = mockFpsMonitor ?? FpsMonitor.instance;
    fpsMonitor.stopScene(previousPage);
    fpsMonitor.startScene(page);

    // fmp 计算终止
    final APMPageTimeMonitor fpmPageTimeMonitor =
        mockApmPageTimeMonitor ?? APMPageTimeMonitor.instance;
    fpmPageTimeMonitor.stopFCPFMP(previousPage);

    if (page == UNKNOWN) {
      return;
    }

    num pageDidPushTime = new DateTime.now().millisecondsSinceEpoch;
    // pv
    final AppContext appContext = mockAppContext ?? AppContext.instance;
    appContext.pageLaunch(page, route.hashCode, pageDidPushTime);
    // fp
    WidgetsBinding.instance?.addPostFrameCallback((callback) {
      appContext.pageFpFinished(route.hashCode);
    });

    // FCP&FMP
    Duration routeTransitionDuration = Duration.zero;
    if (route is TransitionRoute) {
      routeTransitionDuration = route.transitionDuration;
    }
    fpmPageTimeMonitor.monitorFCPFMP(
        page, route.hashCode, routeTransitionDuration, previousRoute == null);

    Timeline.finishSync();
  }

  @override
  void didPop(Route route, Route? previousRoute) {
    FlutterError.resetErrorCount();
    _currentRoute = previousRoute;
    _recordRoute(_RouteAction.POP, route, previousRoute);

    String page = getRouteName(route);
    String previousPage = getRouteName(previousRoute);
    // fps
    final FpsMonitor fpsMonitor = mockFpsMonitor ?? FpsMonitor.instance;
    fpsMonitor.stopScene(page);
    fpsMonitor.startScene(previousPage);
    // FCP&FMP
    final APMPageTimeMonitor fpmPageTimeMonitor =
        mockApmPageTimeMonitor ?? APMPageTimeMonitor.instance;
    fpmPageTimeMonitor.stopFCPFMP(page);
    if (page == UNKNOWN) {
      return;
    }
    final PageLoadDataManager pageLoadDataManager =
        mockPageLoadDataManager ?? PageLoadDataManager.instance;
    pageLoadDataManager.reportAllPageLoadData();
  }

  @override
  void didRemove(Route route, Route? previousRoute) {
    if (_currentRoute == route) {
      FlutterError.resetErrorCount();
      _currentRoute = previousRoute;
    }
    _recordRoute(_RouteAction.REMOVE, route, previousRoute);
  }

  @override
  void didReplace({Route? newRoute, Route? oldRoute}) {
    FlutterError.resetErrorCount();
    _currentRoute = newRoute;
    _recordRoute(_RouteAction.REPLACE, newRoute, oldRoute);
  }

  void _recordRoute(_RouteAction action, Route? route1, Route? route2) {
    if (routeHistory.length == maxRouteHistory) {
      routeHistory.removeLast();
    }
    String route1Name = getRouteName(route1);
    String route2Name = getRouteName(route2);
    String recordHistory = '';
    String actionStr = '';
    switch(action) {
      case _RouteAction.PUSH:
        actionStr = 'push';
        if (routeHistory.length == 0 && route2Name == UNKNOWN) {
          // 第一个页面的previousRoute为null，它的routeName为UNKNOW，该信息不上报
          recordHistory = '打开页面($route1Name)';
        } else {
          recordHistory = '页面($route2Name)跳转到页面($route1Name)';
        }
        break;
      case _RouteAction.POP:
        actionStr = 'pop';
        recordHistory = '页面($route1Name)返回到页面($route2Name)';
        break;
      case _RouteAction.REPLACE:
        actionStr = 'replace';
        recordHistory = '页面($route1Name)替换掉页面($route2Name)';
        break;
      case _RouteAction.REMOVE:
        actionStr = 'remove';
        recordHistory = '移除掉页面($route1Name)'; //route2Name总是为根节点，该信息不上报
        break;
    }
    routeHistory.addFirst(DateTime.now().toIso8601String() + "  " + recordHistory);
    if (kDebugMode) {
      print("action: $actionStr, route1: $route1Name, route2: $route2Name");
    }
  }
}

enum _RouteAction {
  PUSH,
  POP,
  REPLACE,
  REMOVE
}
