import 'dart:async';
import 'dart:math';
import 'dart:ui';

import 'package:ve_apm/src/perf/app_context.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/utils/log.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

typedef void OnFCP();
typedef void OnFMP();

const double xCoverageThreshold = 0.5;
const double yCoverageThreshold = 0.45;
const int fmpTimeOutMilliseconds = 3000;

/// fmp检测开关关闭并且fcp检测完成时，返回的尺寸
class _FmpCloseAndFcpDoneRect extends Rect {
  _FmpCloseAndFcpDoneRect.fromLTRB() : super.fromLTRB(0, 0, 0, 0);
}

class APMPageTimeMonitor {
  static APMPageTimeMonitor? _instance;

  static APMPageTimeMonitor get instance =>
      _instance ??= APMPageTimeMonitor._internal(
          screenWidth: window.physicalSize.width / window.devicePixelRatio,
          screenHeight: window.physicalSize.height / window.devicePixelRatio);

  APMPageTimeMonitor._internal(
      {required double screenWidth, required double screenHeight})
      : this._screenRect = Rect.fromLTWH(0, 0, screenWidth, screenHeight),
        this._areaThreshold = screenWidth * xCoverageThreshold *
            screenHeight * yCoverageThreshold {
    _onFCP = () {
      if (_isFCPing) {
        AppContext.instance.pageFcpFinished(_loadRouteHashCode);
        _isFCPing = false;
      }
    };
  }

  final double _areaThreshold;
  final Rect _screenRect;
  late OnFCP _onFCP;

  /// 页面是否正在加载
  bool _isPageLoading = false;
  bool _isFCPing = false;
  bool _needCheck = true;
  Timer? _frameTimeOutTimer;
  /// 转场动画延时之后fmp检查超时计时器
  Timer? _checkFMPTimeOuterTimer;

  /// 加载的页面信息
  String? _loadPage;
  int _loadRouteHashCode = -1;
  RenderObject? _currentPageRootNode;

  /// 最大的fmp耗时
  int maxFMPTime = 8000;
  int startCheckTime = 0;

  void stopFCPFMP(String? pageName) {
    if (_loadPage == pageName) {
      _reset();
    }
  }

  /// 页面push时，开始监听FCP和FMP
  void monitorFCPFMP(String page,
      int routeHashCode,
      Duration routeTransitionDuration,
      bool isFirstPage,) {
    if (page.isEmpty) {
      return;
    }

    _needCheck = true;
    _isPageLoading = true;
    _isFCPing = true;

    _loadPage = page;
    _loadRouteHashCode = routeHashCode;

    startCheckTime = DateTime.now().millisecondsSinceEpoch;

    if (isFirstPage) {
      // 首页没有转场动画
      WidgetsBinding.instance?.addPostFrameCallback((timeStamp) {
        _checkFCPFMP(routeHashCode);
      });
    } else {
      _checkFMPTimeOuterTimer?.cancel();
      // 非首页则需要跳过转场动画，动画期间不做FCP、FMP检查
      Future.delayed(routeTransitionDuration, () {
        // 增加超时检查逻辑，防止转场动画延时结束后没有新的帧回调
        final num frameCallBackTime = DateTime.now().millisecondsSinceEpoch;
        _checkFMPTimeOuterTimer = Timer(Duration(milliseconds: fmpTimeOutMilliseconds), () {
          _reportFMP(
            'TimeOut',
            fmpTime: frameCallBackTime,
          );
        });
        // 帧回调监听
        WidgetsBinding.instance?.addPostFrameCallback((timeStamp) {
          _checkFMPTimeOuterTimer?.cancel();
          _checkFMPTimeOuterTimer = null;
          _checkFCPFMP(routeHashCode);
        });
      });
    }
  }

  void _reset() {
    try {
      _currentPageRootNode = null;
      _isPageLoading = false;
      _isFCPing = false;
      _needCheck = false;

      _frameTimeOutTimer?.cancel();
      _frameTimeOutTimer = null;

      _checkFMPTimeOuterTimer?.cancel();
      _checkFMPTimeOuterTimer = null;

      _loadPage = null;
      _loadRouteHashCode = -1;
    } catch (e) {}
  }

  void _reportFMP(String type, {num fmpTime = -1}) {
    final int loadRouteHashCode = _loadRouteHashCode;
    _reset();
    if (Switcher.fmpSwitch) {
      AppContext.instance.pageFmpFinished(
        loadRouteHashCode,
        type,
        fmpTime: fmpTime,
      );
    }
  }

  void _checkFCPFMP(int routeHashCode) {
    if (!_isPageLoading || !_needCheck) {
      return;
    }
    // 判断当前帧回调是否为最新注册（避免新旧注册的帧回调重复执行）
    if (_loadRouteHashCode != routeHashCode) {
      return;
    }

    // 兜底逻辑，如果这个时间内还在计算中，则强制停止计算
    final int now = DateTime.now().millisecondsSinceEpoch;
    if (now - startCheckTime >= maxFMPTime) {
      _reportFMP('TimeOut', fmpTime: now);
      return;
    }

    final int startTime = DateTime
        .now()
        .microsecondsSinceEpoch;

    _isPageLoading = _isPageStillLoading(
      renderRoot: WidgetsBinding.instance!.renderView,
      onFMP: () => _reportFMP('Coverage'),
    );
    // String loadPage = _loadPage ?? 'unknown';
    // logProfile(
    //     'page [$loadPage] loading fmp check cost time : ${DateTime
    //         .now()
    //         .microsecondsSinceEpoch - startTime}μs');
    if (_isPageLoading) {
      // 增加超时机制， 若三秒没有帧回调，则认为可交互
      num frameCallBackTime = DateTime.now().millisecondsSinceEpoch;
      _frameTimeOutTimer =
          Timer(Duration(milliseconds: fmpTimeOutMilliseconds), () {
            _reportFMP(
              'TimeOut',
              fmpTime: frameCallBackTime,
            );
          });

      //触发下一帧检测
      WidgetsBinding.instance?.addPostFrameCallback((timeStamp) {
        _frameTimeOutTimer?.cancel();
        _frameTimeOutTimer = null;
        _checkFCPFMP(routeHashCode);
      });
    }
  }

  /// 发送用户点击时，认为可交互
  void onPointDown() {
    if (_isPageLoading && !_isFCPing &&
        (_loadPage?.isNotEmpty ?? false)) {
      _reportFMP('PointDown');
    }
  }

  /// 当前页面是否还在加载。返回true，表示还在加载；返回false，表示加载完成
  ///
  /// 该方法做3件事：
  /// 1. 调用[_findCurrentPageRootNode]找到RenderObject树中，代表当前页面根节点的RenderObject：[_currentPageRootNode]
  /// 2. 调用[_coverElection]算出[_currentPageRootNode]树中内容节点占用的Rect大小
  /// 3. 计算Rect面积（宽*高）是否达到FMP条件的阈值[_areaThreshold]，达到阈值的话回调FMP Callback，并返回false表示页面加载完成。否则返回true表示页面还在加载
  bool _isPageStillLoading({
    required RenderView renderRoot,
    required OnFMP onFMP,
  }) {
    try {
      if (renderRoot.child != null) {
        final Rect? coverRect = _coverElection(renderRoot.child!);
        if (coverRect is _FmpCloseAndFcpDoneRect) {
          return false;
        }
        if (!(coverRect?.isEmpty ?? true) && coverRect!.width * coverRect.height >= _areaThreshold) {
          onFMP();
          return false;
        }
      } else {
        logDebug('current page root is null, traverse view tree in next frame callback.');
      }
    } catch (e) {
      logDebug(e);
    }
    return true;
  }

  /// 计算以[renderObject]为根节点的RenderObject树的内容覆盖尺寸Rect。
  /// 1. 若出现屏幕覆盖率达标情况 直接返回屏幕尺寸，减少计算
  /// 2. 叶子内容节点：返回节点的尺寸
  /// 3. 单孩子节点：孩子尺寸不为空，则认为该节点全覆盖。返回该节点的尺寸
  /// 4. 多孩子节点：计算所有孩子的尺寸，然后调用[_getRectOfContainerRenderObject]算出该节点的尺寸（类美国选举算法）
  Rect? _coverElection(RenderObject renderObject, {
    AxisDirection? parentAxisDirection,
  }) {
    if (renderObject.paintBounds.isEmpty) {
      return null;
    }

    // 叶子节点：文字-RenderParagraph; 图片-RenderImage; 视频-TextureBox
    if (renderObject is RenderParagraph ||
        renderObject is RenderImage ||
        renderObject is TextureBox) {
      // 存在这些Render，则认为是有效屏
      _onFCP();
      if (!Switcher.fmpSwitch) {
        return _FmpCloseAndFcpDoneRect.fromLTRB();
      }
      // 返回当前叶子节点的尺寸
      final Rect renderRect = _getRectRelativeParent(renderObject, parentAxisDirection);
      if (renderRect.width * renderRect.height >= _areaThreshold) {
        return _screenRect;
      }
      return renderRect;
    } else if (renderObject is ContainerRenderObjectMixin) {
      // 多孩子节点
      final ContainerRenderObjectMixin containerRenderObject = renderObject;
      final List<RenderObject> childRenderList = <RenderObject>[];
      containerRenderObject.visitChildrenForSemantics((child) {
        childRenderList.add(child);
      });

      final List<Rect> rectList = <Rect>[];
      for (final RenderObject child in childRenderList) {
        final Rect? childRect = _coverElection(
          child,
          parentAxisDirection: _getAxisDirection(containerRenderObject),
        );
        if (childRect is _FmpCloseAndFcpDoneRect) {
          // 如果关闭fmp检测，并且fcp检测完成，则提前返回
          return childRect;
        }
        if (childRect == _screenRect) {
          return _screenRect;
        }
        if (childRect != null) {
          rectList.add(childRect);
        }
      }
      return _getRectOfContainerRenderObject(
          rectList, containerRenderObject, parentAxisDirection);
    } else if (renderObject is RenderObjectWithChildMixin) {
      // 单孩子节点
      final RenderObjectWithChildMixin renderObjectWithChild = renderObject;
      if (renderObjectWithChild.child != null) {
        final Rect? childRect = _coverElection(
          renderObjectWithChild.child!,
          parentAxisDirection: _getAxisDirection(renderObjectWithChild),
        );
        if (childRect is _FmpCloseAndFcpDoneRect) {
          // 如果关闭fmp检测，并且fcp检测完成，则提前返回
          return childRect;
        }
        if (childRect != null && !childRect.isEmpty) {
          if (childRect == _screenRect) {
            return _screenRect;
          }
          // 孩子尺寸不为空，则认为该RenderObject全覆盖。返回该RenderObject的Rect
          return _getRectRelativeParent(renderObjectWithChild, parentAxisDirection);
        }
      }
    }
    return null;
  }

  /// 类美国选举算法（胜者全赢，败者全输）:
  /// 调用[_mergeChildrenRect]对[rectList]进行合并算出childrenRect，与[renderObject]的面积进行比较
  /// 如果覆盖率达标，则认为[renderObject]是全覆盖，直接返回[renderObject]的Rect；
  /// 否则[renderObject]没有全覆盖，直接返回null。
  ///
  /// 覆盖率达标条件:
  /// childrenRect面积 >= [renderObject]面积 * [xCoverageThreshold] * [yCoverageThreshold]
  Rect? _getRectOfContainerRenderObject(List<Rect> rectList, RenderObject renderObject,
      AxisDirection? parentAxisDirection) {
    if (rectList.isEmpty) {
      return null;
    }

    final Rect renderObjectRect = _getRectRelativeParent(renderObject, parentAxisDirection);
    final Rect childrenRect = _mergeChildrenRect(rectList);
    if (childrenRect.width * childrenRect.height >=
        renderObjectRect.width * renderObjectRect.height *
            xCoverageThreshold * yCoverageThreshold) {
      return renderObjectRect;
    }
    return null;
  }

  /// 合并[rectList]为一个Rect。
  /// 合并算法：把所有Rect的宽、高线段进行投影合并，得出合并后的宽高
  Rect _mergeChildrenRect(List<Rect> rectList) {
    final List<Offset> xProjectionLines = <Offset>[];
    final List<Offset> yProjectionLines = <Offset>[];
    rectList.forEach((rect) {
      if (!rect.isEmpty) {
        xProjectionLines.add(Offset(rect.left, rect.right));
        yProjectionLines.add(Offset(rect.top, rect.bottom));
      }
    });

    final double xProjection = _mergeProjection(xProjectionLines);
    final double yProjection = _mergeProjection(yProjectionLines);
    return Rect.fromCenter(center: Offset(0, 0), width: xProjection, height: yProjection);
  }

  // 计算这些线段的总投影长度
  double _mergeProjection(List<Offset> lines) {
    // Offset.dx表示线段开始，Offset.dy表示线段结束
    if (lines.isEmpty) {
      return 0.0;
    }

    // 按start排序
    lines.sort((first, second) => first.dx.compareTo(second.dx));

    double projectionLength = 0.0;

    double start = -1.0;
    double maxEnd = -1.0;
    for (Offset line in lines) {
      if (maxEnd < line.dx) {
        if (maxEnd > start) {
          projectionLength += maxEnd - start;
        }
        start = line.dx;
      }
      maxEnd = max(maxEnd, line.dy);
    }
    if (maxEnd > start) {
      projectionLength += maxEnd - start;
    }

    return projectionLength;
  }

  /// 相对父布局的Offset
  Offset _getOffsetRelativeParent(RenderObject renderObject,
      AxisDirection? parentAxisDirection) {
    Offset offset = Offset.zero;
    try {
      parentAxisDirection ??= AxisDirection.down;

      final ParentData? parentData = renderObject.parentData;
      if (parentData is BoxParentData) {
        final BoxParentData boxParentData = parentData;
        offset = boxParentData.offset;
      } else if (parentData is SliverGridParentData) {
        final SliverGridParentData sliverGridParentData = parentData;
        final double layoutOffset = sliverGridParentData.layoutOffset ?? 0;
        final double crossAxisOffset =
            sliverGridParentData.crossAxisOffset ?? 0;
        offset = (parentAxisDirection == AxisDirection.down ||
            parentAxisDirection == AxisDirection.up)
            ? Offset(crossAxisOffset, layoutOffset)
            : Offset(layoutOffset, crossAxisOffset);
      } else if (parentData is SliverPhysicalParentData) {
        final SliverPhysicalParentData sliverPhysicalParentData = parentData;
        offset = sliverPhysicalParentData.paintOffset;
      } else if (parentData is SliverLogicalParentData) {
        final SliverLogicalParentData sliverLogicalParentData = parentData;
        final double layoutOffset = sliverLogicalParentData.layoutOffset ?? 0;
        offset = (parentAxisDirection == AxisDirection.down ||
            parentAxisDirection == AxisDirection.up)
            ? Offset(0, layoutOffset)
            : Offset(layoutOffset, 0);
      }
    } catch (e) {
      logDebug(e);
    }
    return offset;
  }

  /// 在父布局坐标系中的Rect
  Rect _getRectRelativeParent(RenderObject renderObject,
      AxisDirection? parentAxisDirection) {
    try {
      Offset offset = _getOffsetRelativeParent(renderObject, parentAxisDirection);
      if (renderObject is RenderBox) {
        final RenderBox renderBox = renderObject;
        return offset & renderBox.size;
      } else {
        return renderObject.paintBounds.shift(offset);
      }
    } catch (e) {
      logDebug(e);
    }
    return renderObject.paintBounds;
  }

  AxisDirection _getAxisDirection(RenderObject renderObject) {
    AxisDirection axisDirection = AxisDirection.down;
    try {
      // ignore: invalid_use_of_protected_member
      final Constraints constraints = renderObject.constraints;
      if (constraints is SliverConstraints) {
        axisDirection = constraints.axisDirection;
      }
    } catch (e) {
      logDebug(e);
    }
    return axisDirection;
  }
}
