import 'dart:async';

import 'package:ve_apm/src/core/apm_navigator_observer.dart';
import 'package:ve_apm/src/core/apm_binding.dart';
import 'package:ve_apm/src/core/apm_widgets_binding_observer.dart';
import 'package:ve_apm/src/errortrace/apm_error_trace.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';
import 'package:ve_apm/src/network/apm_network_error.dart';
import 'package:ve_apm/src/perf/app_context.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:flutter/widgets.dart';
import 'package:ve_apm/src/utils/log.dart';

// abstract class TraceCallback {
//   ///执行真正的runApp逻辑
//   ///需将observer传递给MaterialApp的navigatorObservers
//   ///这样才能完成路由的监听
//   void runRealApp(NavigatorObserver observer);

//   /// 包含初始化参数， 仅针对不接原生监控库场景(目前仅限于desktop和Web版本)
//   Map<String, String> getInitParams() {
//     return {};
//   }

//   bool enableReplaceDefaultErrorWidget() {
//     return true;
//   }
// }

void runTraceApp(FutureOr<Widget> Function(NavigatorObserver) builder,
    {bool enableReplaceDefaultErrorWidget = true}) {
  runZonedGuarded(() async {
    assert(WidgetsBinding.instance == null, 'Pelease call runOnekitApp() first in main().');

    // for monitor app launch fp cost, start
    AppContext.instance.launchStartTime = DateTime.now().millisecondsSinceEpoch;

    // 创建根widget，和OneKit结合使用时，包括了WidgetsFlutterBinding的初始化
    final Widget rootWidget = await builder(ApmNavigatorObserver.instance);

    // 单独使用APM时初始化WidgetsFlutterBinding
    ApmWidgetsFlutterBinding.ensureInitialized();

    final _origErrorHandler = FlutterError.onError;
    FlutterError.onError = (FlutterErrorDetails details) {
      logRelease('method: runTraceApp, step: Zone.current.handleUncaughtError, param: $details');
      Zone.current.handleUncaughtError(details, StackTrace.current);
      _origErrorHandler?.call(details);
    };

    if (enableReplaceDefaultErrorWidget) {
      ApmErrorTrace.instance.replaceErrorWidget();
    }
    // for monitor app launch fp cost, end
    WidgetsBinding.instance!.addPostFrameCallback((callback) {
      AppContext.instance.appFpFinished();
    });
    // for monitor auto fps
    WidgetsBinding.instance!.addObserver(ApmWidgetsBindingObserver());
    // execute biz code
    runApp(rootWidget);

    // AppContext.instance.traceCallback = traceCallback;

    ApmInsight.init();

    // for report backend params(only valid in desktop and web)
    // Map<String, String> initParams = traceCallback.getInitParams();
    // Apm.init(initParams[KEY_AID]);

    // flutter session monitor
    logRelease('method: runAppTrace, step: FpmReporter.reportEvent, param: ${Switcher.SESSION.value}');
    ApmReporter.instance.reportEvent(Switcher.SESSION.value);
  }, (exception, stack) {
    if (exception is ApmNetworkError) {
      // 与原生对齐：网络异常不走异常上报，不参与应用的异常率计算。
      // 网络异常在发生异常[_HttpRecorderImpl#onError]时,就已经上报过了。所以这里直接返回。
      logRelease('method: runTraceApp, step: ApmNetworkError return');
      return;
    }
    FlutterErrorDetails error;
    if (exception is FlutterErrorDetails) {
      error = exception;
      if (error.stack == null) {
        error = error.copyWith(stack: stack);
      }
    } else {
      error = FlutterErrorDetails(
        exception: exception,
        stack: stack,
        context: ErrorDescription(exception.toString()),
      );
    }
    logRelease('method: runTraceApp, step: ApmErrorTrace.recordError, param: $error');
    ApmErrorTrace.instance.recordError(error);
  });
}
