import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';
import 'package:ve_apm/src/utils/log.dart';

enum InterceptType {
  SEND,               //通过send方法拦截
  PLATFORM_MESSAGE,   //通过handlePlatformMessage方法拦截
}

enum IdentityType {
  SYSTEM,  //系统channel
  USER,    //用户channel
}

class ChannelInfo {
  ChannelInfo({
    required this.channelName,
    required this.interceptType,
    required this.startTimestamp,
    this.methodName,
    this.inputSize = 0,
    this.outputSize = 0,
  });

//<editor-fold desc="同步后台的数据" defaultstate="collapsed">
  final String channelName;
  final InterceptType interceptType;
  final int startTimestamp;
  late IdentityType identifyType;
  late int timeCost;
  int inputSize;
  int outputSize;
  String? methodName;

  int get _interceptType {
    int result = 0;
    switch(interceptType) {
      case InterceptType.SEND:
        result = 0;
        break;
      case InterceptType.PLATFORM_MESSAGE:
        result = 1;
        break;
    }
    return result;
  }

  int get _identifyType {
    int result = 0;
    switch(identifyType) {
      case IdentityType.SYSTEM:
        result = 0;
        break;
      case IdentityType.USER:
        result = 1;
        break;
    }
    return result;
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is ChannelInfo &&
          runtimeType == other.runtimeType &&
          channelName == other.channelName &&
          interceptType == other.interceptType &&
          identifyType == other.identifyType &&
          methodName == other.methodName &&
          timeCost == other.timeCost &&
          inputSize == other.inputSize &&
          outputSize == other.outputSize;

  @override
  int get hashCode =>
      channelName.hashCode ^
      interceptType.hashCode ^
      identifyType.hashCode ^
      methodName.hashCode ^
      timeCost.hashCode ^
      inputSize.hashCode ^
      outputSize.hashCode;

  Map<String, dynamic> toMap() {
    final Map<String, dynamic> result = <String, dynamic>{
      'ch': channelName,
      'act': timeCost,
      'isz': inputSize,
      'osz': outputSize,
      'itt': _interceptType,
      'iyt': _identifyType,
    };
    if (methodName != null) {
      result['mth'] = methodName;
    }
    return result;
  }
}

class ChannelInfoManager {

  ChannelInfoManager._();

  static ChannelInfoManager? _instance;

  static ChannelInfoManager get instance => _instance ??= ChannelInfoManager._();

  final int _queueSize = 20;
  final List<Map<String, dynamic>> _queue = <Map<String, dynamic>>[];
  bool _isUploading = false;
  Timer? _reportTimer;

  @visibleForTesting
  Timer? get reportTimer => _reportTimer;

  @visibleForTesting
  int get dataSize => _queue.length;

  @visibleForTesting
  ApmReporter? mockApmReporter;

  @visibleForTesting
  void clearData() => _queue.clear();

  void enqueue(ChannelInfo channelInfo) {
    if (_queue.isEmpty) {
      _reportTimer = Timer.periodic(const Duration(minutes: 2), (Timer timer) {
        _upload();
      });
    }
    _queue.add(channelInfo.toMap());
    if (_queue.length >= _queueSize) {
      _upload();
    }
  }

  /// 聚合上报
  void _upload() {
    if (_queue.isEmpty || _isUploading) {
      return;
    }
    _isUploading = true;
    ApmReporter apmReporter = mockApmReporter ?? ApmReporter.instance;
    logRelease('method: ChannelInfoManager._upload, step: FpmReporter.reportEvent');
    apmReporter.reportEvent(
        Switcher.CHANNEL.value, metric: <String, dynamic>{
      'channels': _queue.toList(),
    });
    _queue.clear();
    _reportTimer?.cancel();
    _reportTimer = null;
    _isUploading = false;
  }

}