import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:flutter/cupertino.dart';

import 'apm_network_data_manager.dart';
import 'apm_network_error.dart';

abstract class IHttpRecorder {
  void onStart(String originUri, String method, String callStackTrace);

  void onRequest(HttpClientRequest httpClientRequest);

  void onResponse(HttpClientResponse response);

  Exception onError(dynamic error, [StackTrace? stackTrace]);
}

typedef HttpRecorderBuilder = IHttpRecorder Function();

class ApmHttpOverrides extends HttpOverrides {
  ApmHttpOverrides({this.customHttpOverrides, 
    @visibleForTesting this.mockHttpRecorder,
    @visibleForTesting this.mockNetworkDataReporter,
    @visibleForTesting this.mockOriginClient});

  HttpOverrides? customHttpOverrides;
  
  @visibleForTesting
  HttpRecorderBuilder? mockHttpRecorder;

  @visibleForTesting
  NetworkDataReporter? mockNetworkDataReporter;

  @visibleForTesting
  HttpClient? mockOriginClient;

  @override
  HttpClient createHttpClient(SecurityContext? context) {
    HttpRecorderBuilder httpRecorderBuilder = mockHttpRecorder ??
        () => _HttpRecorderImpl(mockNetworkDataReporter: mockNetworkDataReporter);
    if (customHttpOverrides != null) {
      return _ApmHttpClient(
          customHttpOverrides!.createHttpClient(context), httpRecorderBuilder);
    }
    // 如果没有定义customHttpOverrides，则先把HttpOverrides.global置空，防止递归调用，使得_HttpClient(默认的HttpClient)可以被初始化
    HttpOverrides.global = null;
    final HttpClient clientWrapper =
        _ApmHttpClient(mockOriginClient ?? HttpClient(context: context), httpRecorderBuilder);
    // HttpClient创建好之后，HttpOverrides.global赋值为原值
    HttpOverrides.global = this;
    return clientWrapper;
  }
}

class _HttpRecorderImpl extends IHttpRecorder {

  _HttpRecorderImpl({@visibleForTesting this.mockNetworkDataReporter});

  @visibleForTesting
  NetworkDataReporter? mockNetworkDataReporter;

  final NetworkDataItem _networkDataItem = NetworkDataItem.empty();

  @override
  void onStart(String originUri, String method, String callStackTrace) {
    _networkDataItem.uri = originUri;
    _networkDataItem.method = method;
    _networkDataItem.callStackTrace = callStackTrace;
    _networkDataItem.startTimestamp = DateTime.now().millisecondsSinceEpoch;
  }

  @override
  void onRequest(HttpClientRequest httpClientRequest) {
    _networkDataItem.request = httpClientRequest;
    _networkDataItem.requestTimestamp = DateTime.now().millisecondsSinceEpoch;
  }

  @override
  void onResponse(HttpClientResponse response) {
    _networkDataItem.response = response;
    _networkDataItem.responseTimestamp = DateTime.now().millisecondsSinceEpoch;
    final NetworkDataReporter networkDataReporter =
        mockNetworkDataReporter ?? NetworkDataReporter.instance;
    if (response.statusCode >= 200 && response.statusCode < 400) {
      networkDataReporter.reportNetworkEvent(NetworkLogTye.SUCCESS_REQUEST, _networkDataItem);
    } else {
      // 状态码异常时，实际上并没有发生exception，所以调用reportError时，不传递exception和stackTrace
      networkDataReporter.reportNetworkEvent(
          NetworkLogTye.STATUS_CODE_ERROR, _networkDataItem,
          reason: 'http request error statusCode.');
    }
  }

  @override
  Exception onError(dynamic error, [StackTrace? stackTrace]) {
    var exception = error?.toString() ?? '';
    if (error is Error) {
      exception += '\n${error.stackTrace}';
    }
    // 请求异常直接上报
    final NetworkDataReporter networkDataReporter =
        mockNetworkDataReporter ?? NetworkDataReporter.instance;
    networkDataReporter.reportNetworkEvent(
        NetworkLogTye.REQUEST_ERROR, _networkDataItem,
        exception: exception, stackTrace: StackTrace.current.toString());
    return ApmNetworkError(error);
  }
}

class _ApmHttpClient extends __HttpClientWrapper {
  _ApmHttpClient(HttpClient origin, this.httpRecorderBuilder) : super(origin);

  HttpRecorderBuilder httpRecorderBuilder;

  @override
  Future<HttpClientRequest> monitorRequest(
      {required String originUri,
      required String method,
      required Future<HttpClientRequest> requestFuture}) async {
    String callStackTrace = StackTrace.current.toString();
    IHttpRecorder httpRecorder = httpRecorderBuilder();
    httpRecorder.onStart(originUri, method, callStackTrace);
    final HttpClientRequest request =
        await requestFuture.onError((error, stackTrace) {
      // 异常还是继续抛出，便于在开发阶段发现异常
      throw httpRecorder.onError(error, stackTrace);
    });
    httpRecorder.onRequest(request);
    return _ApmHttpClientRequest(request, httpRecorder);
  }
}

class _ApmHttpClientRequest extends __HttpClientRequestWrapper {
  _ApmHttpClientRequest(HttpClientRequest origin, this.httpRecorder)
      : super(origin);

  IHttpRecorder httpRecorder;

  @override
  Future<HttpClientResponse> close() async {
    final HttpClientResponse response =
        await origin.close().onError((error, stackTrace) {
      // 异常还是继续抛出，便于在开发阶段发现异常
      throw httpRecorder.onError(error, stackTrace);
    });
    httpRecorder.onResponse(response);
    return response;
  }
}

abstract class __HttpClientWrapper implements HttpClient {
  __HttpClientWrapper(this.origin);

  final HttpClient origin;

  Future<HttpClientRequest> monitorRequest(
      {required String originUri,
      required String method,
      required Future<HttpClientRequest> requestFuture});

  @override
  set autoUncompress(bool value) => origin.autoUncompress = value;

  @override
  bool get autoUncompress => origin.autoUncompress;

  @override
  set idleTimeout(Duration value) => origin.idleTimeout = value;

  @override
  Duration get idleTimeout => origin.idleTimeout;

  @override
  set connectionTimeout(Duration? value) => origin.connectionTimeout = value;

  @override
  Duration? get connectionTimeout => origin.connectionTimeout;

  @override
  set maxConnectionsPerHost(int? value) => origin.maxConnectionsPerHost = value;

  @override
  int? get maxConnectionsPerHost => origin.maxConnectionsPerHost;

  @override
  set userAgent(String? value) => origin.userAgent = value;

  @override
  String? get userAgent => origin.userAgent;

  @override
  void addCredentials(
      Uri url, String realm, HttpClientCredentials credentials) {
    origin.addCredentials(url, realm, credentials);
  }

  @override
  void addProxyCredentials(
      String host, int port, String realm, HttpClientCredentials credentials) {
    origin.addProxyCredentials(host, port, realm, credentials);
  }

  @override
  set authenticate(
      Future<bool> Function(Uri url, String scheme, String realm)? f) {
    origin.authenticate = f;
  }

  @override
  set authenticateProxy(
      Future<bool> Function(String host, int port, String scheme, String realm)?
      f) {
    origin.authenticateProxy = f;
  }

  @override
  set badCertificateCallback(
      bool Function(X509Certificate cert, String host, int port)? callback) {
    origin.badCertificateCallback = callback;
  }

  @override
  void close({bool force = false}) {
    origin.close(force: force);
  }

  @override
  set findProxy(String Function(Uri url)? f) {
    origin.findProxy = f;
  }

  String _buildUrl(String host, int port, String path) {
    return 'http://$host:$port/$path';
  }

  @override
  Future<HttpClientRequest> delete(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'DELETE',
        requestFuture: origin.delete(host, port, path));
  }

  @override
  Future<HttpClientRequest> deleteUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'DELETE',
        requestFuture: origin.deleteUrl(url));
  }

  @override
  Future<HttpClientRequest> get(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'GET',
        requestFuture: origin.get(host, port, path));
  }

  @override
  Future<HttpClientRequest> getUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'GET',
        requestFuture: origin.getUrl(url));
  }

  @override
  Future<HttpClientRequest> head(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'HEAD',
        requestFuture: origin.head(host, port, path));
  }

  @override
  Future<HttpClientRequest> headUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'HEAD',
        requestFuture: origin.headUrl(url));
  }

  @override
  Future<HttpClientRequest> open(
      String method, String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: method,
        requestFuture: origin.open(method, host, port, path));
  }

  @override
  Future<HttpClientRequest> openUrl(String method, Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: method,
        requestFuture: origin.openUrl(method, url));
  }

  @override
  Future<HttpClientRequest> patch(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'PATCH',
        requestFuture: origin.patch(host, port, path));
  }

  @override
  Future<HttpClientRequest> patchUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'PATCH',
        requestFuture: origin.patchUrl(url));
  }

  @override
  Future<HttpClientRequest> post(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'POST',
        requestFuture: origin.post(host, port, path));
  }

  @override
  Future<HttpClientRequest> postUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'POST',
        requestFuture: origin.postUrl(url));
  }

  @override
  Future<HttpClientRequest> put(String host, int port, String path) {
    return monitorRequest(
        originUri: _buildUrl(host, port, path),
        method: 'PUT',
        requestFuture: origin.put(host, port, path));
  }

  @override
  Future<HttpClientRequest> putUrl(Uri url) {
    return monitorRequest(
        originUri: '${url.toString()}',
        method: 'PUT',
        requestFuture: origin.putUrl(url));
  }
}

abstract class __HttpClientRequestWrapper implements HttpClientRequest {
  __HttpClientRequestWrapper(this.origin);

  final HttpClientRequest origin;

  @override
  bool get bufferOutput => origin.bufferOutput;

  @override
  set bufferOutput(bool value) => origin.bufferOutput = value;

  @override
  int get contentLength => origin.contentLength;

  @override
  set contentLength(int value) => origin.contentLength = value;

  @override
  Encoding get encoding => origin.encoding;

  @override
  set encoding(Encoding value) => origin.encoding = value;

  @override
  bool get followRedirects => origin.followRedirects;

  @override
  set followRedirects(bool value) => origin.followRedirects = value;

  @override
  int get maxRedirects => origin.maxRedirects;

  @override
  set maxRedirects(int value) => origin.maxRedirects = value;

  @override
  set persistentConnection(bool value) => origin.persistentConnection = value;

  @override
  bool get persistentConnection => origin.persistentConnection;

  @override
  HttpHeaders get headers => origin.headers;

  @override
  String get method => origin.method;

  @override
  Uri get uri => origin.uri;

  @override
  HttpConnectionInfo? get connectionInfo => origin.connectionInfo;

  @override
  List<Cookie> get cookies => origin.cookies;

  @override
  void write(Object? obj) {
    origin.write(obj);
  }

  @override
  void writeAll(Iterable<dynamic> objects, [String separator = '']) {
    origin.writeAll(objects, separator);
  }

  @override
  void writeCharCode(int charCode) {
    origin.writeCharCode(charCode);
  }

  @override
  void writeln([Object? obj = '']) {
    origin.writeln(obj);
  }

  @override
  void add(List<int> data) {
    origin.add(data);
  }

  @override
  void addError(Object error, [StackTrace? stackTrace]) {
    origin.addError(error, stackTrace);
  }

  @override
  Future<dynamic> addStream(Stream<List<int>> stream) {
    return origin.addStream(stream);
  }

  @override
  Future<HttpClientResponse> get done => origin.done;

  @override
  Future<dynamic> flush() {
    return origin.flush();
  }

  @override
  void abort([Object? exception, StackTrace? stackTrace]) {
    return origin.abort(exception, stackTrace);
  }
}
