import 'dart:async';
import 'dart:io';

import 'package:flutter/widgets.dart';
import 'package:ve_apm/src/errortrace/apm_error_trace.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/utils/log.dart';
import 'package:ve_apm/ve_apm.dart';
import 'apm_network_error.dart';

/// 网络请求需要上报的数据段
class NetworkDataItem {
  NetworkDataItem.empty();

  late String uri;
  late String method;
  late String callStackTrace;
  int startTimestamp = -1; //开始请求的时间戳

  HttpClientRequest? request;
  int requestTimestamp = -1; //request创建完成的时间戳

  HttpClientResponse? response;
  int responseTimestamp = -1; //response返回完成的时间戳

  Map<String, dynamic> get requestInfo => _requestInfo();

  Map<String, dynamic> _requestInfo() {
    Map<String, dynamic> result = <String, dynamic>{
      'uri': uri,
      'method': method,
    };
    if (request != null) {
      result['uri'] = request!.uri.toString();
      result['request_size'] = request!.headers.contentLength + request!.contentLength;
      result['method'] = request!.method;
    }
    if (response != null) {
      result['response_size'] = response!.headers.contentLength + response!.contentLength;
      result['total_spend_time'] = responseTimestamp - requestTimestamp;
      result['status_code'] = response!.statusCode;
      List<Map> redirects = [];
      response!.redirects.forEach((element) {
        Map<String, dynamic> redirectInfo = <String, dynamic>{
          'code': element.statusCode,
          'method': element.method,
          'uri': element.location.origin,
        };
        redirects.add(redirectInfo);
      });
      if (redirects.isNotEmpty) {
        result['redirect_info'] = redirects;
      }
    }
    return result;
  }
}

class NetworkDataReporter {

  static NetworkDataReporter? _instance;

  static NetworkDataReporter get instance =>
      _instance ??= NetworkDataReporter._internal();

  NetworkDataReporter._internal();

  /// 成功的请求会缓存起来上报；异常的请求会直接上报
  void reportNetworkEvent(NetworkLogTye logTye, NetworkDataItem networkDataItem,
      {String? exception,
      String? reason,
      String? information,
      String? stackTrace,
      @visibleForTesting dynamic mockSwitcher,
      @visibleForTesting ApmReporter? mockApmReporter}) {
    bool networkSwitcher;
    if (mockSwitcher != null) {
      networkSwitcher =
          mockSwitcher.isModuleReportable(Switcher.NETWORK) as bool;
    } else {
      networkSwitcher = Switcher.isModuleReportable(Switcher.NETWORK);
    }
    if (logTye == NetworkLogTye.SUCCESS_REQUEST && !networkSwitcher) {
      // 如果是成功的请求并且没命中采样，那么不进行上报，方法提前返回；
      // 如果是异常请求，无论采样是否命中，都需要上报
      return;
    }
    final Map<String, dynamic> metric = <String, dynamic>{
      "log_type": logTye.value,
      "network_switcher": networkSwitcher ? 1 : 0, //表示是否命中采样
      "request_info": networkDataItem.requestInfo,
    };
    if (logTye != NetworkLogTye.SUCCESS_REQUEST) {
      final Map<String, dynamic> errorInfo = ApmErrorTrace.instance
          .generateBaseErrorInfo(
              exception: exception,
              reason: reason,
              information: information,
              stackTrace: stackTrace);
      errorInfo[KEY_CALL_STACKTRACE] = networkDataItem.callStackTrace;
      metric['error_info'] = errorInfo;
    }
    if (kDebugMode) {
      logDebug('network monitor log: $metric');
    }
    ApmReporter apmReporter = mockApmReporter ?? ApmReporter.instance;
    logRelease('method: NetworkDataReporter, step: FpmReporter.reportEvent, param: ${Switcher.NETWORK.value};$metric');
    apmReporter.reportEvent(Switcher.NETWORK.value, metric: metric);
  }
}
