import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:ve_apm/src/core/constants.dart';
import 'package:ve_apm/src/core/apm_navigator_observer.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';
import 'package:ve_apm/src/perf/time_tracer.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/utils/log.dart';
import 'package:ve_apm/ve_apm.dart';

typedef FpsStopCallback = void Function(double averageFps);

/// 累计, 上报前求平均
class _FpsCalculator {
  _FpsCalculator(this.scene);
  final String scene;
  int count = 0;

  double fpsSum = 0;
  double uiAvgTimeSum = 0;
  double gpuAvgTimeSum = 0;

  int wallSum = 0; // ms

  /// 用于计算 uiFps 和 丢帧率, 所有不需要就平均
  /// 表示丢 key 帧的情况，发生了 value 次
  Map<int, int> uiSkipped = <int, int>{};

  bool isValid() => count > 0 && uiSkipped.isNotEmpty;

  num get uiFps {
    int skippedCount = 0;
    int drawnCount = 0;
    uiSkipped.forEach((skipped, cnt) {
      drawnCount += cnt;
      skippedCount += skipped * cnt;
    });
    final int totalCount = drawnCount + skippedCount;
    if (totalCount > 0) {
      return TimeTracer.defaultDisplayRefreshRate * drawnCount / totalCount;
    } else {
      return TimeTracer.defaultDisplayRefreshRate;
    }
  }

  int get duration {
    assert(isValid());
    return wallSum ~/ count;
  }

  void addData(FpsData data) {
    assert(scene == data.scene);
    if (data.isValid()) {
      count += 1;
      wallSum += data.wall;
      data.skippedMap?.forEach((k, v) {
        if (uiSkipped.containsKey(k)) {
          uiSkipped[k] = uiSkipped[k]! + v;
        } else {
          uiSkipped[k] = v;
        }
      });
    }
  }
}

/// FPS 采集器，开始采集时调用 [startScene]，停止采集时调用 [stopScene]
/// 默认情况下，框架会在route push 和 pop自动进行 start和stop, 所以当统计页面级的FPS时，无需手动调用startScene 和 StopScene
/// 内部注册了 WidgetsBindingObserver， 在 resumed 和 paused 状态时，分别调用  [onAppLifecycleResumed] 和 [onAppLifecyclePaused]
class FpsMonitor {
  FpsMonitor._internal();

  static FpsMonitor? _instance;

  static FpsMonitor get instance => _instance ??= FpsMonitor._internal();

  /// 所有通过 [startScene] 静态方法创建的 monitors
  static final Map<String, SceneFpsMonitor> _tracingMonitors = <String, SceneFpsMonitor>{};

  @visibleForTesting
  Map<String, SceneFpsMonitor> get tracingMonitors => _tracingMonitors;
  
  void startScene(String scene, {Duration? timeout}) {
    if (scene == UNKNOWN) {
      return;
    }
    if (_tracingMonitors.containsKey(scene)) {
      return;
    }
    final SceneFpsMonitor monitor = SceneFpsMonitor(scene);
    monitor.start(timeout: timeout);
    _tracingMonitors[scene] = monitor;
  }

  void stopScene(String scene) {
    if (scene == UNKNOWN) {
      return;
    }
    final SceneFpsMonitor? monitor = _tracingMonitors.remove(scene);
    monitor?.stop();
  }

  void onAppLifecycleResumed() {
    startScene(ApmNavigatorObserver.instance.getCurrentRouteName());
  }

  void onAppLifecyclePaused() {
    stopScene(ApmNavigatorObserver.instance.getCurrentRouteName());
  }
  
}

/// 具体Scene的FpsMonitor
class SceneFpsMonitor {
  /// [sceneName]用来区分上报的场景
  SceneFpsMonitor(this.sceneName,
      {@visibleForTesting TimeTracer? mockTimeTracer,
      @visibleForTesting this.mockFpsReporter}) {
    _tracer = mockTimeTracer ?? TimeTracer(sceneName: sceneName);
  }
  
  static bool debug = false;
  static bool enableSkipFrame = true;

  late TimeTracer _tracer;
  Timer? _timer;

  /// 上报的场景信息
  final String sceneName;

  @visibleForTesting
  FpsReporter? mockFpsReporter;

  @visibleForTesting
  Timer? get timer => _timer;

  /// 开始采集，一般在开始滑动时调用
  /// 这里的超时 定义为：调用start后，stop前，每 timeout 时间会取一次丢帧序列
  void start({Duration? timeout}) {
    if (SceneFpsMonitor.debug) {
      print('FpsMonitor start');
    }
    if (_tracer.isTracing) {
      return;
    }
    _timer?.cancel();
    _timer = Timer.periodic(timeout ??= const Duration(seconds: 30),
            (Timer timer) => _onTime());
    _tracer.start();
  }

  void _onTime() {
    if (SceneFpsMonitor.debug) {
      print('FpsMonitor onTime');
    }
    if (!_tracer.isTracing) {
      return;
    }
    final FpsData? data = _tracer.onTime();
    if (data?.isValid() ?? false) {
      final FpsReporter fpsReporter = mockFpsReporter ?? FpsReporter.instance;
      fpsReporter.enqueue(data!);
    }
  }

  /// 结束采集，一般在滑动结束时调用
  void stop() {
    if (SceneFpsMonitor.debug) {
      print('FpsMonitor stop');
    }
    if (!_tracer.isTracing) {
      return;
    }
    if (_timer != null && _timer!.isActive) {
      _timer!.cancel();
      _timer = null;
    }

    final FpsData? data = _tracer.stop();
    if (data?.isValid() ?? false) {
      final FpsReporter fpsReporter = mockFpsReporter ?? FpsReporter.instance;
      fpsReporter.enqueue(data!);
    }
  }
}

class FpsReporter {
  FpsReporter._internal();

  static FpsReporter? _instance;

  static FpsReporter get instance => _instance ??= FpsReporter._internal();

  bool _init = false;

  /// 当前使用的上报模式
  Timer? _timer;
  final Map<String?, _FpsCalculator> _calculators =
      <String?, _FpsCalculator>{};

  @visibleForTesting
  int get calculatorsLength => _calculators.length;

  @visibleForTesting
  void clear() => _calculators.clear();

  /// 初始化 monitor
  void ensureInitialized() {
    if (_init) {
      return;
    }
    _init = true;
    _timer = Timer.periodic(const Duration(minutes: 2), _timerCallBack);
  }

  /// 销毁监控数据
  void destroy() {
    if (SceneFpsMonitor.debug) {
      print('destroy, cancel timer');
    }
    if (_timer != null) {
      _timer!.cancel();
      _timer = null;
    }
    flushAll();
    _calculators.clear();
  }

  /// 添加 FPS 信息到List
  void enqueue(FpsData fpsData) {
    ensureInitialized();
    if (SceneFpsMonitor.debug) {
      print('FpsReport enqueue: ${fpsData.scene},'
          ' wall=${fpsData.wall}ms'
          ' skipped=${fpsData.skippedMap}');
    }
    if (!_calculators.containsKey(fpsData.scene)) {
      _calculators[fpsData.scene] = _FpsCalculator(fpsData.scene);
    }
    _calculators[fpsData.scene]!.addData(fpsData);
  }

  /// 上报 FPS 数据的timer
  void _timerCallBack(Timer timer) {
    flushAll();
  }

  void flushAll({@visibleForTesting ApmReporter? mockApmReporter}) {
    if (SceneFpsMonitor.debug) {
      print('FpsReporter flushAll scenes=${_calculators.keys}');
    }
    _calculators.forEach((String? scene, _FpsCalculator calculator) {
      if (!calculator.isValid()) {
        return;
      }
      // json map 的 key 必须是 string
      final Map<String, int> _uiSkipped = <String, int>{};
      calculator.uiSkipped.forEach((k, v) {
        _uiSkipped['$k'] = v;
      });

      final Map<String, dynamic> metric = <String, dynamic>{
        // 'duration_ms': calculator.duration,
        'fps': calculator.uiFps,
      };
      final Map<String, dynamic> category = <String, dynamic>{'scene': scene};
      // TODO, refresh rate needs to support dynamic query
      final Map<String, dynamic> extra = <String, dynamic>{'skipped': _uiSkipped, 'hz': TimeTracer.defaultDisplayRefreshRate};
      if (SceneFpsMonitor.debug) {
        print('monitorEvent fps: sceneName=$scene metirc=$metric extra=$extra');
      }
      final ApmReporter apmReporter = mockApmReporter ?? ApmReporter.instance;
      logRelease('method: FpsMonitor.flushAll, step: ApmReporter.innerUseReportEventNoPublic, param:$category;$metric;$extra');
      apmReporter.innerUseReportEventNoPublic(Switcher.FPS.value, category: category, metric: metric, extra: extra);
    });

    _calculators.clear();
  }
}
