import 'dart:async';
import 'dart:collection';

import 'package:flutter/material.dart';
import 'package:ve_apm/src/core/constants.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/utils/log.dart';

// 页面加载需要上报的数据
class PageLoadData {
  String page;
  num pageDidPushTime;
  num? fp;
  num? fcp;
  String? fmpType;
  num? fmp;

  PageLoadData({this.page = UNKNOWN, this.pageDidPushTime = -1, this.fp, this.fcp, this.fmp});
}

typedef UpdatePageLoadDataFun = void Function(PageLoadData pageLoadData);

class PageLoadDataManager {
  PageLoadDataManager._internal();

  static PageLoadDataManager? _instance;

  static PageLoadDataManager get instance =>
      _instance ??= PageLoadDataManager._internal();

  //一个page多次加载时，page是一样的，route是不一样的。因此以pageRouteHashCode作为缓存的key，而非page。
  final LinkedHashMap<int, PageLoadData> _pageLoadDataCache = LinkedHashMap();
  Timer? _reportTimer;

  @visibleForTesting
  LinkedHashMap<int, PageLoadData> get pageLoadDataCache => _pageLoadDataCache;

  @visibleForTesting
  Timer? get reportTimer => _reportTimer;

  @visibleForTesting
  ApmReporter? mockApmReporter;

  /// 相同的pageRouteHashCode会覆盖数据，如果要更新数据，请调用[updatePageLoadData]
  void joinPageLoadData(int pageRouteHashCode, String page,
      num pageDidPushTime) {
    if (_pageLoadDataCache.isEmpty) {
      _reportTimer = Timer.periodic(const Duration(seconds: 30), (Timer timer) {
        _reportPageLoadDataExceptLastPage();
      });
    }
    final PageLoadData value = PageLoadData(page: page, pageDidPushTime: pageDidPushTime);
    _pageLoadDataCache[pageRouteHashCode] = value;
  }

  void updatePageLoadData(int pageRouteHashCode, UpdatePageLoadDataFun updatePageLoadDataFun) {
    final PageLoadData? value = _pageLoadDataCache[pageRouteHashCode];
    if (value != null) {
      updatePageLoadDataFun.call(value);
    }
  }

  /// 调用时机：1.pop事件；2.App Pause事件
  void reportAllPageLoadData() {
    _pageLoadDataCache.forEach((key, value) {
      _reportPageLoadData(value);
    });
    _pageLoadDataCache.clear();
    _reportTimer?.cancel();
    _reportTimer = null;
  }

  /// 计时器上报时调用。计时器上报场景下，最后一个页面的数据可能还没收集完，下一次再上报
  /// 计时器是为了解决"一直push页面没有触发pop事件；并且没有切换到后台触发App Pause事件" 场景下的上报问题
  void _reportPageLoadDataExceptLastPage() {
    List<int> keys = _pageLoadDataCache.keys.toList();
    keys.removeLast();
    keys.forEach((element) {
      var value = _pageLoadDataCache[element]!;
      _reportPageLoadData(value);
      _pageLoadDataCache.remove(element);
    });
  }

  void _reportPageLoadData(PageLoadData value) {
    final Map<String, num> metric = {};
    if (value.fp != null) {
      metric['fp'] = value.fp!;
    }
    if (value.fcp != null) {
      metric['fcp'] = value.fcp!;
    }
    if (value.fmp != null) {
      metric['fmp'] = value.fmp!;
    }
    Map<String, dynamic>? extra = <String, dynamic>{};
    if (value.fmpType != null) {
      extra['fmpType'] = value.fmpType!;
    }
    extra['fmp_switch'] = Switcher.fmpSwitch ? 1 : 0;
    final ApmReporter apmReporter = mockApmReporter ?? ApmReporter.instance;
    logRelease('method: PageLoadDataManager._reportPageLoadData, step: FpmReporter.reportEvent, param:$metric;$extra');
    apmReporter.reportEvent(
      Switcher.PAGE_LOAD.value,
      timestamp: value.pageDidPushTime,
      category: <String, dynamic>{'page_key': value.page},
      metric: metric,
      extra: extra,
    );
  }
}
