import 'dart:ui';

import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';

class FpsData {
  FpsData(this.scene, this.wall, this.skippedMap);
  final String scene;
  final int wall; // ms
  /// 丢帧表: key 表示丢了几帧，value 表示发生的次数
  final Map<int, int>? skippedMap;

  bool isValid() => skippedMap != null && skippedMap!.isNotEmpty;

  @override
  String toString() {
    return 'FpsData{_scene: $scene, wall: $wall, skippedMap: $skippedMap}';
  }
}

class TimeTracer {
  TimeTracer({required this.sceneName, @visibleForTesting Stopwatch? mockStopWatch})
      : _stopwatch = mockStopWatch ?? Stopwatch();
  static bool registered = false;
  static late List<int> frameMapGlobal;

  static const int kFrameMapSize = 600; // record 599 skipped frames at most (10s)
  static const int defaultDisplayRefreshRate = 60;

  String sceneName;
  List<int>? _start;
  List<int>? _current;
  bool _tracing = false;
  final Stopwatch _stopwatch;

  /// wall time 墙上时间, ms
  int get wall => _stopwatch.elapsedMilliseconds;
  
  bool get isTracing => _tracing;

  @visibleForTesting
  List<int>? get startFrameMap => _start;

  @visibleForTesting
  List<int>? get currentFrameMap => _current;


  static void ensureRegistered() {
    if (!registered) {
      frameMapGlobal = List<int>.filled(kFrameMapSize, 0, growable: false);
      const double frameIntervalMS =
          1000 / defaultDisplayRefreshRate; // 16.6ms per frame
      final TimingsCallback callback = (List<FrameTiming> timings) {
        for (final FrameTiming frameTiming in timings) {
          final int skippedFrameCount =
               frameTiming.totalSpan.inMilliseconds ~/ frameIntervalMS;
          if (skippedFrameCount < frameMapGlobal.length) {
            frameMapGlobal[skippedFrameCount]++;
          }
        }
      };
      SchedulerBinding.instance!.addTimingsCallback(callback);
      registered = true;
    }
  }

  void start() {
    if (_tracing) {
      return;
    }
    _tracing = true;
    ensureRegistered();
    _stopwatch.reset();
    _stopwatch.start();
    _start = List<int>.from(frameMapGlobal);
  }

  FpsData? onTime() {
    if (!_tracing) {
      return null;
    }
    _stopwatch.stop();
    final Map<int, int> skippedMap = _getSkippedMap();
    final int fpsWall = wall;
    _start = _current;
    _resumeStopwatch();
    return FpsData(sceneName, fpsWall, skippedMap);
  }

  void _resumeStopwatch() {
    _stopwatch.start();
  }

  FpsData? stop() {
    if (!_tracing) {
      return null;
    }
    _tracing = false;
    _stopwatch.stop();
    final Map<int, int> skippedMap = _getSkippedMap();
    final int fpsWall = wall;
    _start = null;
    _current = null;
    return FpsData(sceneName, fpsWall, skippedMap);
  }

  /// 丢帧表: key 表示丢了几帧，value 表示发生的次数
  Map<int, int> _getSkippedMap() {
    _current = List<int>.from(frameMapGlobal);
    final Map<int, int> skippedMap = <int, int>{};
    for (int i = 0; i < _current!.length; i++) {
      final int diff = _current![i] - _start![i];
      if (diff > 0) {
        skippedMap[i] = diff;
      }
    }
    return skippedMap;
  }
}