import 'dart:collection';
import 'dart:convert';

import 'package:flutter/cupertino.dart';
import 'package:flutter/services.dart';

import '../utils/log.dart';

class ApmInsight {
  static const _MAX_CACHES = 300;
  static const MethodChannel _channel = const MethodChannel('ve_apm');
  static List<_ReportNode> _crashCache = [];
  static Queue<_ReportNode> _eventCache = Queue();
  static bool _isInitialized = false;

  static void init({@visibleForTesting MethodChannel? mockChannel}) async {
    if (_isInitialized) {
      return;
    }
    _isInitialized = true;
    MethodChannel channel = mockChannel ?? _channel;
    channel.setMethodCallHandler(_handleMethodCall);
    // 主动获取一次配置。避免插件预注册时，执行了configUpdate，但是Flutter#main方法没执行，
    // 无法响应configUpdate导致后续flutterConfig一直为空。
    String? flutterConfig = await channel.invokeMethod<String>('getFlutterConfig');
    if (flutterConfig != null) {
      Map config = json.decode(flutterConfig) as Map;
      Switcher.updateReportableModules(config);
    }
  }

  static void reportToInsight(Map<String, dynamic> eventMap,
      {@visibleForTesting MethodChannel? mockChannel}) {
    assert(_isInitialized == true, 'must call ApmInsight.init() first');
    final String? module = eventMap['service'] as String?;
    if (module == null) {
      logRelease('method: ApmInsight.reportToInsight, step: module is null return');
      return;
    }
    final eventString = jsonEncode(eventMap);
    Map<String, dynamic> map = <String, dynamic>{"log": eventString};
    Switcher switcher = Switcher._(module);
    if (!Switcher.isReady) {
      logRelease('method: ApmInsight.reportToInsight, step: ApmInsight._cacheInsert, param: $map');
      _cacheInsert(switcher, map);
    } else if (Switcher.isModuleReportable(switcher)) {
      logRelease('method: ApmInsight.reportToInsight, step: invock reportToInsight, param: $map');
      MethodChannel channel = mockChannel ?? _channel;
      channel.invokeMethod<void>("reportToInsight", map);
    }
  }

  static void reportDartError(Map<String, dynamic> eventMap,
      {@visibleForTesting MethodChannel? mockChannel}) {
    assert(_isInitialized == true, 'must call ApmInsight.init() first');
    if (eventMap.isEmpty) {
      logRelease('method: reportToDartError, step: eventMap is empty return');
      return;
    }
    final eventString = jsonEncode(eventMap);
    Map<String, dynamic> map = <String, dynamic>{"log": eventString};
    if (!Switcher.isReady) {
      logRelease('method: reportToDartError, step: ApmInsight._cacheInsert, param: $map');
      _cacheInsert(Switcher.CRASH, map);
    } else if (Switcher.isModuleReportable(Switcher.CRASH)) {
      logRelease('method: reportToDartError, step: invoke reprotDartError, param: $map');
      MethodChannel channel = mockChannel ?? _channel;
      channel.invokeMethod<void>("reportDartError", map);
    }
  }

  static Future<dynamic> _handleMethodCall(MethodCall call,
      {@visibleForTesting MethodChannel? mockChannel}) async {
    if ("configUpdate" == call.method) {
      if (call.arguments is String) {
        Map config = json.decode(call.arguments as String) as Map;
        Switcher.updateReportableModules(config);
        _cacheClean(mockChannel: mockChannel);
      }
    }
  }

  static void _cacheInsert(Switcher switcher, Map<String, dynamic> params) {
    if (Switcher.CRASH == switcher) {
      _crashCache.add(_ReportNode(Switcher.CRASH, params));
    } else {
      if (_eventCache.length >= _MAX_CACHES) {
        _eventCache.removeFirst();
      }
      _eventCache.add(_ReportNode(switcher, params));
    }
  }

  static void _cacheClean({@visibleForTesting MethodChannel? mockChannel}) {
    MethodChannel channel = mockChannel ?? _channel;
    if (Switcher.isReady) {
      if (Switcher.isModuleReportable(Switcher.CRASH)) {
        _crashCache.forEach((node) {
          channel.invokeMethod<void>("reportDartError", node.params);
        });
      }
      _crashCache.clear();
      _eventCache.forEach((node) {
        if (Switcher.isModuleReportable(node.switcher)) {
          channel.invokeMethod<void>("reportToInsight", node.params);
        }
      });
      _eventCache.clear();
    }
  }

  @visibleForTesting
  static Future<dynamic> handleMethodCall(MethodCall call,
          {@visibleForTesting MethodChannel? mockChannel}) =>
      _handleMethodCall(call, mockChannel: mockChannel);

  @visibleForTesting
  static List<_ReportNode> get crashCache => _crashCache;

  @visibleForTesting
  static Queue<_ReportNode> get eventCache => _eventCache;
}

class _ReportNode {
  Switcher switcher;
  Map<String, dynamic> params;
  _ReportNode(this.switcher, this.params);
  String toString() => "${switcher.value}: $params";
}

/// 与后端下发的采样开关Config相对应
/// 如果增加了采样指标，需要更新这个类：增加新的Switcher实例，
/// 并为[updateReportableModules]方法的增加switch case分支，记录新增指标的开关情况
class Switcher {

  const Switcher._(this._value);

  /// 调用[reportEvent(eventName)]方法时，eventName可以传入Switcher实例的value值，对应不同模块的上报
  String get value => _value;
  final String _value;

  static const Switcher CRASH = Switcher._('flutter_crash');
  static const Switcher SESSION = Switcher._('flutter_session');
  static const Switcher PAGE_LOAD = Switcher._('flutter_page_load');
  static const Switcher FMP = Switcher._('flutter_fmp');
  static const Switcher FPS = Switcher._('flutter_fps');
  static const Switcher NETWORK = Switcher._('flutter_network');
  static const Switcher CHANNEL = Switcher._('flutter_channel');
  static const List<Switcher> _ITEMS = [CRASH, SESSION, PAGE_LOAD, FMP, FPS, NETWORK, CHANNEL];
  static Map<String, bool> _reportableModules = {}; //采样率配置

  static bool get isReady => _reportableModules.isNotEmpty;

  // 暂时通过这个接口暴露给外部，可用于控制是否进行数据收集
  static bool isModuleReportable(Switcher switcher) =>
      isReady && (_reportableModules[switcher.value] ?? false);

  //fmp检测开关: isReady为false时，表示config还没下发。这时默认采集fmp
  static bool get fmpSwitch => isModuleReportable(Switcher.FMP) || !isReady;

  /// config格式：
  /// {
  ///     "flutter_fps_module": {
  ///       "switcher": 1,
  ///       "flutter_fps_ratio": 1 //1表示命中，0表示未命中
  ///     },
  ///     "flutter_crash_module":{
  ///       "switcher": 1,
  ///     },
  ///     "flutter_page_load_module":{
  ///       "switcher": 1,
  ///       "flutter_page_load_ratio": 1,
  ///       "flutter_fmp_ratio": 0,
  ///     },
  ///     "flutter_network_module": {
  ///       "switcher": 1,
  ///       "flutter_network_ratio": 1,
  ///     }
  ///   },
  static void updateReportableModules(Map config) {
    bool isModuleReportable(String moduleName, String ratioName) {
      final Map moduleConfig = (config['${moduleName}_module'] as Map?) ?? <dynamic, dynamic>{};
      return 1 == moduleConfig["switcher"] && 1 == (moduleConfig['${ratioName}_ratio'] ?? 1);
    }
    _ITEMS.forEach((element) {
      bool result = false;
      switch (element) {
        case Switcher.CRASH:
        case Switcher.PAGE_LOAD:
        case Switcher.FPS:
        case Switcher.NETWORK:
        case Switcher.CHANNEL:
          result = isModuleReportable(element.value, element.value);
          break;
        case Switcher.SESSION:
          //服务端没有下发session的上报配置，和crash上报配置保持一致
          result = isModuleReportable(Switcher.CRASH.value, Switcher.CRASH.value);
          break;
        case Switcher.FMP:
          result = isModuleReportable(Switcher.PAGE_LOAD.value, Switcher.FMP.value);
          break;
      }
      _reportableModules[element.value] = result;
    });
  }

  @visibleForTesting
  static void clear() {
    _reportableModules.clear();
  }
}