import 'dart:typed_data';
import 'dart:ui';

import 'package:flutter/services.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:ve_apm/src/core/apm_binding.dart';
import 'package:ve_apm/src/core/channel_info_manager.dart';

import 'apm_binding_test.mocks.dart';

class TestBinaryMessenger extends BinaryMessenger {
  TestBinaryMessenger(
      {this.handlePlatformMessageCallbackData, this.sendReturnData});

  final ByteData? handlePlatformMessageCallbackData;
  final ByteData? sendReturnData;

  @override
  bool checkMessageHandler(String channel, MessageHandler? handler) {
    return false;
  }

  @override
  bool checkMockMessageHandler(String channel, MessageHandler? handler) {
    return false;
  }

  @override
  Future<void> handlePlatformMessage(String channel, ByteData? data,
      PlatformMessageResponseCallback? callback) async {
    callback?.call(handlePlatformMessageCallbackData);
  }

  @override
  Future<ByteData?>? send(String channel, ByteData? message) async {
    return sendReturnData;
  }

  @override
  void setMessageHandler(String channel, MessageHandler? handler) {}

  @override
  void setMockMessageHandler(String channel, MessageHandler? handler) {}
}

@GenerateMocks([ChannelInfoManager])
void main() {
  group('ApmBinaryMessenger class test', () {
    tearDownAll(() {
      ApmBinaryMessenger.mockChannelInfoManager = null;
    });

    test('handlePlatformMessage: callback无论是否是null，都会记录outputSize和timeCost',
        () {
      const String channelName = 'test_channel';
      const String methodName = 'test_method';
      final ByteData inputData = const StandardMethodCodec()
          .encodeMethodCall(const MethodCall(methodName, {'a': 1}));
      final ByteData outputData =
          ByteData.view(Uint16List.fromList([1]).buffer);
      final int inputSize = inputData.lengthInBytes;
      final int outputSize = outputData.lengthInBytes;
      final mockChannelInfoManager = MockChannelInfoManager();
      ApmBinaryMessenger.mockChannelInfoManager = mockChannelInfoManager;

      void callback_test(PlatformMessageResponseCallback? callback) async {
        final ApmBinaryMessenger messenger = ApmBinaryMessenger(
            TestBinaryMessenger(handlePlatformMessageCallbackData: outputData));
        messenger.handlePlatformMessage(channelName, inputData, callback);
        final ChannelInfo channelInfo =
            verify(mockChannelInfoManager.enqueue(captureAny)).captured.single
                as ChannelInfo;
        expect(channelInfo.channelName, channelName);
        expect(channelInfo.methodName, methodName);
        expect(channelInfo.inputSize, inputSize);
        expect(channelInfo.outputSize, outputSize);
        expect(channelInfo.identifyType, IdentityType.USER);
        expect(channelInfo.interceptType, InterceptType.PLATFORM_MESSAGE);
        expect(channelInfo.timeCost != null, true);
      }

      callback_test((data) => print('callback is not null'));
      callback_test(null);
    });

    test('send: MethodChannel case', () async {
      const String channelName = 'test_channel';
      const String methodName = 'test_method';
      final ByteData inputData = const StandardMethodCodec()
          .encodeMethodCall(const MethodCall(methodName, {'a': 1}));
      final ByteData outputData =
          ByteData.view(Uint16List.fromList([1]).buffer);
      final int inputSize = inputData.lengthInBytes;
      final int outputSize = outputData.lengthInBytes;
      final mockChannelInfoManager = MockChannelInfoManager();
      ApmBinaryMessenger.mockChannelInfoManager = mockChannelInfoManager;

      ApmBinaryMessenger messenger =
          ApmBinaryMessenger(TestBinaryMessenger(sendReturnData: outputData));
      await messenger.send(channelName, inputData);
      final ChannelInfo channelInfo =
          verify(mockChannelInfoManager.enqueue(captureAny)).captured.single
              as ChannelInfo;
      expect(channelInfo.channelName, channelName);
      expect(channelInfo.methodName, methodName);
      expect(channelInfo.inputSize, inputSize);
      expect(channelInfo.outputSize, outputSize);
      expect(channelInfo.identifyType, IdentityType.USER);
      expect(channelInfo.interceptType, InterceptType.SEND);
      expect(channelInfo.timeCost != null, true);
    });

  });
}
