import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:ve_apm/src/core/constants.dart';
import 'package:ve_apm/src/core/apm_navigator_observer.dart';
import 'package:ve_apm/src/core/apm_page_time_monitor.dart';
import 'package:ve_apm/src/perf/app_context.dart';
import 'package:ve_apm/src/perf/fps_monitor.dart';
import 'package:ve_apm/src/perf/page_load_data_manager.dart';

import 'apm_navigator_observer_test.mocks.dart';

@GenerateMocks([FpsMonitor, APMPageTimeMonitor, PageLoadDataManager, AppContext])
void main() {
  group('FpmNavigatorObserver class test', () {
    setUpAll(() {
      TestWidgetsFlutterBinding.ensureInitialized();
    });

    tearDownAll(() {
      ApmNavigatorObserver.instance.setMockObject(
          mockFpsMonitor: null,
          mockApmPageTimeMonitor: null,
          mockPageLoadDataManager: null,
          mockAppContext: null);
    });

    test('push: 首页路由信息为: 打开页面(...)', (){
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;

      const String routeName = '/first_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      const Route? previousRoute = null;
      observer.didPush(route, previousRoute);
      expect(observer.getRecentRouteHistory().first
          .endsWith('打开页面($routeName)'),
          true);
    });

    test('push: 路由信息为: 页面(...)跳转到页面(...)', (){
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;

      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didPush(route, previousRoute);
      expect(observer.getRecentRouteHistory().first
              .endsWith('页面($previousRouteName)跳转到页面($routeName)'),
          true);
    });

    test('pop: 路由信息为: 页面(...)返回到页面(...)', (){
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;

      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didPop(route, previousRoute);
      expect(observer.getRecentRouteHistory().first
          .endsWith('页面($routeName)返回到页面($previousRouteName)'),
          true);
    });

    test('replace: 路由信息为: 页面(...)替换掉页面(...)', (){
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;

      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didReplace(newRoute: route, oldRoute: previousRoute);
      expect(observer.getRecentRouteHistory().first
          .endsWith('页面($routeName)替换掉页面($previousRouteName)'),
          true);
    });

    test('remove: 路由信息为: 移除掉页面(...)', (){
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;

      const String routeName = '/second_page';
      const String previousRouteName = '/first_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didRemove(route, previousRoute);
      expect(observer.getRecentRouteHistory().first
          .endsWith('移除掉页面($routeName)'),
          true);
    });

    test('didPush: 先终止previousRoute的fps收集，后开始currentRoute的fps收集；'
        '先终止previousRoute的fmp计算，后开始currentRoute的fmp计算', () {
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;
      final mockFpsMonitor = MockFpsMonitor();
      final mockApmPageTimeMonitor = MockAPMPageTimeMonitor();
      observer.setMockObject(mockFpsMonitor: mockFpsMonitor, mockApmPageTimeMonitor: mockApmPageTimeMonitor);
      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didPush(route, previousRoute);

      verifyInOrder([
        mockFpsMonitor.stopScene(previousRouteName),
        mockFpsMonitor.startScene(routeName),
      ]);

      verifyInOrder([
        mockApmPageTimeMonitor.stopFCPFMP(previousRouteName),
        mockApmPageTimeMonitor.monitorFCPFMP(routeName, any, any, any)
      ]);
    });

    test('didPush: 打开的页面时UNKNOWN时，不会计算pv、fp、fcp、fmp', () {
      final mockAppContext = MockAppContext();
      final mockApmPageTimeMonitor = MockAPMPageTimeMonitor();
      ApmNavigatorObserver.instance.setMockObject(
          mockApmPageTimeMonitor: mockApmPageTimeMonitor,
          mockAppContext: mockAppContext,
      );
      const String previousRouteName = '/first_page';
      const String routeName = UNKNOWN;
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      ApmNavigatorObserver.instance.didPush(route, previousRoute);
      //触发下一帧，验证fp
      TestWidgetsFlutterBinding.ensureInitialized().scheduleFrame();

      verifyNever(mockAppContext.pageLaunch(any, any, any));
      verifyNever(mockAppContext.pageFpFinished(any));
      verifyNever(mockApmPageTimeMonitor.monitorFCPFMP(any, any, any, any));
    });

    test('didPop: 先终止currentRoute的fps收集，后开始previousRoute的fps收集；终止currentRoute的fmp计算', () {
      ApmNavigatorObserver observer = ApmNavigatorObserver.instance;
      final mockFpsMonitor = MockFpsMonitor();
      final mockApmPageTimeMonitor = MockAPMPageTimeMonitor();
      observer.setMockObject(mockFpsMonitor: mockFpsMonitor, mockApmPageTimeMonitor: mockApmPageTimeMonitor);
      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      observer.didPop(route, previousRoute);
      verifyInOrder([
        mockFpsMonitor.stopScene(routeName),
        mockFpsMonitor.startScene(previousRouteName)
      ]);
      verify(mockApmPageTimeMonitor.stopFCPFMP(routeName)).called(1);

    });

    test('didPop: 退出的页面是UNKNOWN时，不会上报page_load_data', () {
      final mockPageLoadDataManager = MockPageLoadDataManager();
      ApmNavigatorObserver.instance.setMockObject(
        mockPageLoadDataManager: mockPageLoadDataManager,
      );
      const String previousRouteName = '/first_page';
      const String routeName = UNKNOWN;
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      ApmNavigatorObserver.instance.didPush(route, previousRoute);
      //触发下一帧，验证fp
      TestWidgetsFlutterBinding.ensureInitialized().scheduleFrame();

      verifyNever(mockPageLoadDataManager.reportAllPageLoadData());
    });

    test('didPop: 退出的页面非UNKNOWN页面时，上报page_load_data', () {
      final mockPageLoadDataManager = MockPageLoadDataManager();
      ApmNavigatorObserver.instance.setMockObject(
        mockPageLoadDataManager: mockPageLoadDataManager,
      );
      const String previousRouteName = '/first_page';
      const String routeName = '/second_page';
      final Route route = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: routeName));
      final Route previousRoute = MaterialPageRoute<void>(builder: (_) => Container(),
          settings: const RouteSettings(name: previousRouteName));
      ApmNavigatorObserver.instance.didPush(route, previousRoute);
      //触发下一帧，验证fp
      TestWidgetsFlutterBinding.ensureInitialized().scheduleFrame();

      verifyNever(mockPageLoadDataManager.reportAllPageLoadData());
    });

  });


}