import 'dart:io';

import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:pedantic/pedantic.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:ve_apm/src/network/apm_http_overrides.dart';
import 'package:ve_apm/src/network/apm_network_data_manager.dart';
import 'package:ve_apm/src/network/apm_network_error.dart';

import 'apm_http_overrides_test.mocks.dart';

class TestHttpOverrides extends HttpOverrides {
  TestHttpOverrides(this.function);

  Function function;

  @override
  HttpClient createHttpClient(SecurityContext? context) {
    function.call();
    return super.createHttpClient(context);
  }
}

class TestHttpRecorder extends IHttpRecorder {

  static List<int> data = [];
  @override
  Exception onError(dynamic error, [StackTrace? stackTrace]) {
    return Exception();
  }

  @override
  void onRequest(HttpClientRequest httpClientRequest) {
  }

  @override
  void onResponse(HttpClientResponse response) {
    data.add(1);
  }

  @override
  void onStart(String originUri, String method, String callStackTrace) {
  }

}

@GenerateMocks([NetworkDataReporter, HttpClient])
void main() {
  group('ApmHttpOverrides class test', () {
    setUp(startServer);
    tearDown(() {
      stopServer();
      HttpOverrides.global = null;
    });

    test('customHttpOverrides is working, 并且能顺利监控', () async{
      final mockNetworkDataReporter = MockNetworkDataReporter();
      int flag = 0;
      HttpOverrides.global =
          ApmHttpOverrides(customHttpOverrides: TestHttpOverrides(() => flag++),
              mockNetworkDataReporter: mockNetworkDataReporter);
      HttpClient();
      final HttpClient client = HttpClient();
      HttpClientRequest request =
          await client.getUrl(serverUrl.resolve('/success'));
      await request.close();
      verify(mockNetworkDataReporter.reportNetworkEvent(
          NetworkLogTye.SUCCESS_REQUEST, any))
          .called(1);
      expect(flag, 2);
    });

    test('同一个HttpClient进行多次网络请求，会产生多个网络数据', () async{
      HttpOverrides.global = ApmHttpOverrides(mockHttpRecorder: () => TestHttpRecorder());
      final HttpClient client = HttpClient();
      HttpClientRequest request =
          await client.getUrl(serverUrl.resolve('/success'));
      await request.close();

      request = await client.getUrl(serverUrl.resolve('/success'));
      await request.close();

      request = await client.getUrl(serverUrl.resolve('/success'));
      await request.close();

      expect(TestHttpRecorder.data.length, 3);
    });

    test('状态码异常上报', () async {
      final mockNetworkDataReporter = MockNetworkDataReporter();
      HttpOverrides.global =
          ApmHttpOverrides(mockNetworkDataReporter: mockNetworkDataReporter);
      final HttpClient client = HttpClient();
      HttpClientRequest request =
          await client.getUrl(serverUrl.resolve('/404_error'));
      await request.close();
      verify(mockNetworkDataReporter.reportNetworkEvent(
          NetworkLogTye.STATUS_CODE_ERROR, any,
          reason: 'http request error statusCode.'));
    });

    test('传输层异常上报', () async {
      final mockNetworkDataReporter = MockNetworkDataReporter();
      final mockClient = MockHttpClient();
      final uri = Uri.parse('timeout_url');
      when(mockClient.getUrl(uri)).thenAnswer(
          (_) async => throw const SocketException('timeout_error'));
      HttpOverrides.global = ApmHttpOverrides(
          mockNetworkDataReporter: mockNetworkDataReporter,
          mockOriginClient: mockClient);

      final HttpClient client = HttpClient();
      expect(() async => await client.getUrl(uri), throwsA(predicate((e) {
        verify(mockNetworkDataReporter.reportNetworkEvent(
                NetworkLogTye.REQUEST_ERROR, any,
                exception: 'SocketException: timeout_error',
                stackTrace: anyNamed('stackTrace')))
            .called(1);
        return e is ApmNetworkError;
      })));
    });

    test('正常请求上报', () async{
      final mockNetworkDataReporter = MockNetworkDataReporter();
      HttpOverrides.global =
          ApmHttpOverrides(mockNetworkDataReporter: mockNetworkDataReporter);
      final HttpClient client = HttpClient();
      HttpClientRequest request =
          await client.getUrl(serverUrl.resolve('/success'));
      await request.close();
      verify(mockNetworkDataReporter.reportNetworkEvent(
              NetworkLogTye.SUCCESS_REQUEST, any))
          .called(1);
    });
  });
}

/// The current server instance.
HttpServer? _server;

/// The URL for the current server instance.
Uri get serverUrl => Uri.parse('http://localhost:${_server!.port}');

/// Starts a new HTTP server.
Future<void> startServer() async {
  _server = (await HttpServer.bind('localhost', 0))
    ..listen((request) async {
      var path = request.uri.path;
      var response = request.response;

      if (path == '/404_error') {
        response
          ..statusCode = 404
          ..contentLength = 0;
        unawaited(response.close());
        return;
      }

      if (path == '/success') {
        response
          ..statusCode = 200
          ..contentLength = -1
          ..write('body');
        unawaited(response.close());
        return;
      }
    });
}

/// Stops the current HTTP server.
void stopServer() {
  if (_server != null) {
    _server!.close();
    _server = null;
  }
}
