import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:ve_apm/src/core/constants.dart';
import 'package:ve_apm/src/perf/fps_monitor.dart';
import 'package:ve_apm/src/perf/time_tracer.dart';
import 'package:ve_apm/src/plugin/apm_insight.dart';
import 'package:ve_apm/src/report/apm_reporter.dart';

import 'fps_monitor_test.mocks.dart';

@GenerateMocks([TimeTracer, FpsReporter, ApmReporter])
void main() {
  setUpAll(() {
    TestWidgetsFlutterBinding.ensureInitialized();
  });

  group('FpsMonitor class test', () {
    test('startScene: UNKNOWN和存储过的页面会提前返回', () {
      FpsMonitor.instance.tracingMonitors.clear();
      FpsMonitor.instance.startScene(UNKNOWN);
      expect(FpsMonitor.instance.tracingMonitors.isEmpty, true);

      String scene = 'test_page';
      FpsMonitor.instance.startScene(scene);
      final SceneFpsMonitor fpsMonitor =
          FpsMonitor.instance.tracingMonitors[scene]!;
      FpsMonitor.instance.startScene(scene);
      expect(FpsMonitor.instance.tracingMonitors[scene]!, fpsMonitor);
    });

    test('stopScene: UNKNOWN页面会提前返回；', () {
      FpsMonitor.instance.tracingMonitors.clear();
      String scene = 'test_page';
      FpsMonitor.instance.startScene(scene);

      FpsMonitor.instance.stopScene(UNKNOWN);
      expect(FpsMonitor.instance.tracingMonitors.isEmpty, false);

      FpsMonitor.instance.stopScene(scene);
      expect(FpsMonitor.instance.tracingMonitors.isEmpty, true);
    });
  });

  group('SceneFpsMonitor class test', () {
    test('多次调用start，会提前返回', () {
      final mockTimeTracer = MockTimeTracer();
      SceneFpsMonitor sceneFpsMonitor =
          SceneFpsMonitor('test_page', mockTimeTracer: mockTimeTracer);
      when(mockTimeTracer.isTracing).thenReturn(false);
      sceneFpsMonitor.start();
      when(mockTimeTracer.isTracing).thenReturn(true);
      sceneFpsMonitor.start();
      sceneFpsMonitor.start();
      verify(mockTimeTracer.start()).called(1);
      expect(sceneFpsMonitor.timer?.isActive, true);
    });

    test('调用stop，tracer.isTracing为false时，会提前返回', () {
      final mockTimeTracer = MockTimeTracer();
      SceneFpsMonitor sceneFpsMonitor =
          SceneFpsMonitor('test_page', mockTimeTracer: mockTimeTracer);
      when(mockTimeTracer.isTracing).thenReturn(true);
      when(mockTimeTracer.stop()).thenReturn(null);
      sceneFpsMonitor.stop();
      verify(mockTimeTracer.stop()).called(1);
      expect(sceneFpsMonitor.timer == null, true);

      when(mockTimeTracer.isTracing).thenReturn(false);
      sceneFpsMonitor.stop();
      sceneFpsMonitor.stop();
      verifyNever(mockTimeTracer.stop());
    });

    test('start之后，每 timeout 时间会取一次丢帧序列，尝试加入到FpsReporter中', () async {
      final mockTimeTracer = MockTimeTracer();
      final mockFpsReporter = MockFpsReporter();
      final SceneFpsMonitor sceneFpsMonitor = SceneFpsMonitor('test_page',
          mockTimeTracer: mockTimeTracer, mockFpsReporter: mockFpsReporter);
      final fpsData = FpsData('test_page', 1, {0: 1, 1: 1, 2: 1});
      void onTime_test(FpsData? fpsData, int enqueueCallTime) async {
        when(mockTimeTracer.isTracing).thenReturn(false);
        when(mockTimeTracer.onTime()).thenReturn(fpsData);
        sceneFpsMonitor.start(timeout: const Duration(seconds: 1));
        await Future<void>.delayed(const Duration(seconds: 2));
        if (enqueueCallTime == 0) {
          verifyNever(mockFpsReporter.enqueue(fpsData));
        } else {
          verify(mockFpsReporter.enqueue(fpsData)).called(enqueueCallTime);
        }
      }

      onTime_test(fpsData, 2);
      onTime_test(null, 0);
    });

    test('stop之后，会生成一个FpsData，尝试加入到FpsReporter中', () async {
      final mockTimeTracer = MockTimeTracer();
      final mockFpsReporter = MockFpsReporter();
      final SceneFpsMonitor sceneFpsMonitor = SceneFpsMonitor('test_page',
          mockTimeTracer: mockTimeTracer, mockFpsReporter: mockFpsReporter);
      final fpsData = FpsData('test_page', 1, {0: 1, 1: 1, 2: 1});
      void stop_test(FpsData? fpsData, int enqueueCallTime) {
        when(mockTimeTracer.isTracing).thenReturn(false);
        when(mockTimeTracer.stop()).thenReturn(fpsData);
        sceneFpsMonitor.start();
        when(mockTimeTracer.isTracing).thenReturn(true);
        sceneFpsMonitor.stop();
        if (enqueueCallTime == 0) {
          verifyNever(mockFpsReporter.enqueue(fpsData));
        } else {
          verify(mockFpsReporter.enqueue(fpsData)).called(enqueueCallTime);
        }
      }

      stop_test(fpsData, 1);
      stop_test(null, 0);
    });
  });

  group('FpsReporter class test', () {

    tearDownAll((){
      FpsReporter.instance.clear();
    });

    test('enqueue', () {
      final fpsData0 = FpsData('0_page', 1, {0: 1, 1: 1, 2: 1});
      FpsReporter.instance.enqueue(fpsData0);
      expect(FpsReporter.instance.calculatorsLength, 1);

      final fpsData1 = FpsData('1_page', 1, {0: 1, 1: 1, 2: 1});
      FpsReporter.instance.enqueue(fpsData1);
      expect(FpsReporter.instance.calculatorsLength, 2);

      final fpsData11 = FpsData('1_page', 1, {0: 1, 1: 1, 2: 1});
      FpsReporter.instance.enqueue(fpsData11);
      expect(FpsReporter.instance.calculatorsLength, 2);
    });

    test('flushAll', () {
      final fpsData0 = FpsData('0_page', 1, {0: 1, 1: 1, 2: 1});
      FpsReporter.instance.enqueue(fpsData0);
      final fpsData1 = FpsData('1_page', 1, {0: 1, 1: 1, 2: 1});
      FpsReporter.instance.enqueue(fpsData1);
      int calculatorsLength = FpsReporter.instance.calculatorsLength;
      expect(calculatorsLength, 2);

      final mockApmReporter = MockApmReporter();
      FpsReporter.instance.flushAll(mockApmReporter: mockApmReporter);
      verify(mockApmReporter.innerUseReportEventNoPublic(Switcher.FPS.value,
              timestamp: anyNamed('timestamp'),
              category: anyNamed('category'),
              metric: anyNamed('metric'),
              extra: anyNamed('extra')))
          .called(calculatorsLength);
      expect(FpsReporter.instance.calculatorsLength, 0);
    });
  });
}
