import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:ve_apm/src/perf/time_tracer.dart';

import 'time_tracer_test.mocks.dart';

@GenerateMocks([Stopwatch])
void main() {
  setUpAll(() {
    TestWidgetsFlutterBinding.ensureInitialized();
  });

  group('TimeTracer class test', () {
    test('start will call ensureRegistered', () {
      final mockStopWatch = MockStopwatch();
      final TimeTracer timeTracer =
          TimeTracer(sceneName: '', mockStopWatch: mockStopWatch);
      timeTracer.start();
      expect(TimeTracer.registered, true);
      verifyInOrder([mockStopWatch.reset(), mockStopWatch.start()]);
    });

    test('调用start后，isTracing为true；调用stop后，isTracing为false', () {
      final TimeTracer timeTracer = TimeTracer(sceneName: '');
      timeTracer.start();
      expect(timeTracer.isTracing, true);
      timeTracer.stop();
      expect(timeTracer.isTracing, false);
    });

    test('多次调用start，会提前返回', () {
      final mockStopWatch = MockStopwatch();
      final TimeTracer timeTracer =
      TimeTracer(sceneName: '', mockStopWatch: mockStopWatch);
      timeTracer.start();
      timeTracer.start();
      timeTracer.start();
      verify(mockStopWatch.reset()).called(1);
    });

    test('没有调用start时，onTime 和 stop会返回null', () {
      final TimeTracer timeTracer = TimeTracer(sceneName: '');
      expect(timeTracer.onTime(), null);
      expect(timeTracer.stop(), null);
    });

    test('onTime:  Stopwatch will be resumed', () {
      final mockStopWatch = MockStopwatch();
      when(mockStopWatch.elapsedMilliseconds).thenReturn(10);
      final TimeTracer timeTracer =
          TimeTracer(sceneName: '', mockStopWatch: mockStopWatch);
      timeTracer.start();
      verifyInOrder([mockStopWatch.reset(), mockStopWatch.start()]);

      timeTracer.onTime();
      verifyInOrder([
        mockStopWatch.stop(),
        mockStopWatch.elapsedMilliseconds,
        mockStopWatch.start()
      ]);
    });

    test('onTime: _start will be assigned as _current', () async {
      String sceneName ='test_page';
      final TimeTracer timeTracer = TimeTracer(sceneName: sceneName);
      timeTracer.start();
      for (int i in [0, 1, 2, 3]) {
        expect(TimeTracer.frameMapGlobal[i], 0);
      }

      await Future<void>.delayed(const Duration(seconds: 1));
      TimeTracer.frameMapGlobal[0] = 0;
      TimeTracer.frameMapGlobal[1] = 1;
      TimeTracer.frameMapGlobal[2] = 2;
      TimeTracer.frameMapGlobal[3] = 3;
      final FpsData fpsData = timeTracer.onTime()!;
      expect(fpsData.scene, sceneName);
      expect(fpsData.skippedMap, {1: 1, 2: 2, 3: 3});
      for (int i in [0, 1, 2, 3]) {
        expect(timeTracer.startFrameMap![i], i);
        expect(timeTracer.currentFrameMap![i], i);
      }
    });

    test('stop: _start and _current will be null', () async {
      TimeTracer.frameMapGlobal = List<int>.filled(600, 0, growable: false);
      String sceneName ='test_page';
      final TimeTracer timeTracer = TimeTracer(sceneName: sceneName);
      timeTracer.start();
      for (int i in [0, 1, 2, 3]) {
        expect(TimeTracer.frameMapGlobal[i], 0);
      }

      await Future<void>.delayed(const Duration(seconds: 1));
      TimeTracer.frameMapGlobal[0] = 0;
      TimeTracer.frameMapGlobal[1] = 1;
      TimeTracer.frameMapGlobal[2] = 2;
      TimeTracer.frameMapGlobal[3] = 3;
      final FpsData fpsData = timeTracer.stop()!;
      expect(fpsData.scene, sceneName);
      expect(fpsData.skippedMap, {1: 1, 2: 2, 3: 3});
      expect(timeTracer.startFrameMap, null);
      expect(timeTracer.currentFrameMap, null);
    });
  });
}